# -*- encoding: utf-8 -*-
#
# Copyright (c) 2009 Daniel Holth <dholth@fastmail.fm>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

__all__ = ["HEX", "MODHEX", "translate"]

# Possible Yubikey alphabets. Generated by code at
# http://bitbucket.org/dholth/yubikey/
alphabets = u""",yuéebfcstnxg.vi
cbdefghijklnrtuv
cbdefghijklnrtuṣ
cbdefghıjklnrtuv
cbdefghıíklnrtuv
cbdefghıĭklnrtuv
cbdešghijklnrtuv
cbsftdhuneikpglv
cbɗefghijklnrtuv
cžsrtmpuneišldjv
gwcbdthoeaznyusv
jka.oumtdsrqhxlb
jka.oumtdsrĝhĉlb
jxe.iuhcdrnbpygk
jxe.uidchtnbpygk
jxeñuidcrtnbpygk
jxeöuidchtnbpygk
vçexautnkmlziorc
vçeğaütnkmlzıorc
xizqaehutdcn.os,
xkipe,cdtsr'oèv.
¯¦É¿¸À̀¨¾Á¼¶ºÅÃ±
äzaleosgnrtbcwhp
çbdefghijklnrtuý
čñďéëŕúíüôľňřťůç
ŋbdefghijklnrtuv
ŋbdefghiƒklnrtuv
ψβδεφγηιξκλνρτθω
жбдефгңийклнртув
йще.уидцхтнбпыгк
сивуапршолдткегм
сиқуапршолдткегм
сівуапршолдткегм
цбдефгхийклнртув
цбдефгхийклнртуж
цбдефгхијклнртув
цбдефгхійклнртуж
цбдефгчийклнртуж
цбдефґгійклнртув
ъфаеожгстнвхишкэ
ёмбуөахшролижэгс
գբդէֆքհիճկլնրտըվ
գպտէֆկհիճքլնրդըւ
չզգբեանկիտհլսմւյ
ցբդեֆգհիյկլնռտւվ
ցբդեֆգհիյկլնրտւվ
քբդէֆգհիճկլնրտըվ
בנגקכעיןחלךמראוה
ؤﻻيثبلاهتنمىقفعر
جبدەفگهحژکلنرتئڤ
زذیثبلاهتنمدقفعر
چبدعفگحیجکلنرتءط
چبدعفگھیجکلنرتءط
ےشرھنلہباکیغدٹتس
ܤܒܕܖܔܓܗܥܛܟܠܢܪܬܜܫ
ܤܧܝܖܒܠܐܗܬܢܡ܀ܩܦܥܪ
ޗބދެފގހިޖކލނރތުވ
ߗߓߘߍߝߜߤߌߖߞߟߣߙߕߎߢ
चबडेङगहिजकलनरटुव
चबदेटगहिजकलनरतुड
छबदेउगहिजकलनरतुव
मव्ािुपगरकतलीूहन
চবডীতগহিজকলনরটুআ
মব্ািুপগরকতলীূহন
েনিডব্াহকতদসপটজর
ਚਬ੍ਾਿੁਹਗਜਕਲਨੀੂਦਵ
ਮਵ੍ਾਿੁਪਗਰਕਤਲੀੂਹਨ
મવ્ાિુપગરકતલીૂહન
ମଵ୍ାିୁପଗରକତଲୀୂହନ
உெனநகபாைதமடஔசவரஎ
మవ్ాిుపగరకతలీూహన
ಮವ್ಾಿುಪಗರಕತಲೀೂಹನ
ചബദെഫഗഹിജകലനരതുവ
മവ്ാിുപഗരകതലീൂഹന
චබදඑෆගහඉජකලනරතඋව
ลิงยกัีมานเคอรดห
แิกำดเ้ร่าสืพะีอ
ແຶກຳດເ້ຣ່າສືພະີອ
ཀཔདེབངམི་གལནརཏུཁ
འརདགནཔཕོབམཙལངིེཡ
မဗ္ာိုပဂရကတလီူဟန
სივუაპრშოლდტკეგმ
ყჟაუეოდნმსრზძჭთღ
ცბდეფგჰიჯკლნრტუვ
ቸበደeፈገሀiጀከለነረተuሸ
ᏓᎨᏗᎡᎩᎦᎯᎢᏚᎸᎵᎾᏛᏔᎤᎥ
ᖃᑕᖁᕿᑯᑐᓱᓂᒧᓄᓗᓴᑭᑎᒥᑲ
ᚉᚁᚇᚓᚃᚌᚆᚔᚗᚖᚂᚅᚏᚈᚒᚍ
ចបដេថងហិ្កលនរតុវ
ⵛⴱⴷⴻⴼⴳⵀⵉⵊⴽⵍⵏⵔⵜⵓⵖ
ソコシイハキクニマノリミスカナヒ""".split(u"\n")

index = {}
for i, alphabet in enumerate(alphabets):
    for letter in alphabet:
        index.setdefault(letter, set()).update([i])

HEX = u"0123456789abcdef"
MODHEX = u"cbdefghijklnrtuv"


def translate(otp, to=MODHEX):
    """Return set() of possible modhex interpretations of a Yubikey otp.

    If otp uses all 16 characters in its alphabet, there will be only
    one possible interpretation of that Yubikey otp (except for two
    Armenian keyboard layouts).

    otp: Yubikey output.
    to: 16-character target alphabet, default MODHEX.
    """
    if not isinstance(otp, unicode):
        raise ValueError("otp must be unicode")
    if not isinstance(to, unicode):
        raise ValueError("to must be unicode")
    possible = (set(index[c]) for c in set(otp))
    possible = reduce(lambda a, b: a.intersection(b), possible)
    translated = set()
    for i in possible:
        a = alphabets[i]
        translation = dict(zip((ord(c) for c in a), to))
        translated.add(otp.translate(translation))
    return translated



from __future__ import with_statement

import os
import sys
import subprocess
import signal
import time
import socket
import errno
import atexit

from os.path import join as pjoin


def waitForStartUp(process, address, timeout=10):
    # connect to it, with a timeout in case something went wrong
    start = time.time()
    while time.time() < start + timeout:
        try:
            s = socket.create_connection(address)
            s.close()
            break
        except:
            time.sleep(0.1)
    else:
        # see if process is still alive
        process.poll()

        if process and process.returncode is None:
            process.terminate()
        raise RuntimeError("Couldn't connect to server; aborting test")


class ProcessRunner(object):
    def setUp(self, *args, **kwargs):
        pass

    def tearDown(self, *args, **kwargs):
        if self.process:
            self.process.terminate()


class MockAPIServerRunner(ProcessRunner):
    def __init__(self, port=8881):
        self.port = port

    def setUp(self, *args, **kwargs):
        self.cwd = os.getcwd()
        self.process = None
        self.base_dir = pjoin(self.cwd)
        self.log_path = pjoin(self.cwd, 'mock_api_server.log')

        super(MockAPIServerRunner, self).setUp(*args, **kwargs)
        script = pjoin(os.path.dirname(__file__), 'mock_http_server.py')

        with open(self.log_path, 'a+') as log_fp:
            args = '%s --port=%s' % (script, str(self.port))
            args = [script, '--port=%s' % (self.port)]

            self.process = subprocess.Popen(args, shell=False,
                                            cwd=self.base_dir, stdout=log_fp,
                                            stderr=log_fp)
            waitForStartUp(self.process, ('127.0.0.1', self.port), 10)
        atexit.register(self.tearDown)



import sys
import unittest

import requests

from yubico_client import yubico
from yubico_client.otp import OTP
from yubico_client.yubico_exceptions import StatusCodeError
from yubico_client.yubico_exceptions import InvalidClientIdError
from yubico_client.yubico_exceptions import SignatureVerificationError
from yubico_client.yubico_exceptions import InvalidValidationResponse


class TestOTPClass(unittest.TestCase):
    def test_otp_class(self):
        otp1 = OTP('tlerefhcvijlngibueiiuhkeibbcbecehvjiklltnbbl')
        otp2 = OTP('jjjjjjjjnhe.ngcgjeiuujjjdtgihjuecyixinxunkhj',
                   translate_otp=True)

        self.assertEqual(otp1.device_id, 'tlerefhcvijl')
        self.assertEqual(otp2.otp,
                         'ccccccccljdeluiucdgffccchkugjcfditgbglbflvjc')

    def test_translation_multiple_interpretations(self):
        otp_str1 = 'vvbtbtndhtlfguefgluvbdcetnitidgkvfkbicevgcin'
        otp1 = OTP(otp_str1)
        self.assertEqual(otp1.otp, otp_str1)

    def test_translation_single_interpretation(self):
        otp_str1 = 'cccfgvgitchndibrrtuhdrgeufrdkrjfgutfjbnhhglv'
        otp_str2 = 'cccagvgitchndibrrtuhdrgeufrdkrjfgutfjbnhhglv'
        otp1 = OTP(otp_str1)
        otp2 = OTP(otp_str2)
        self.assertEqual(otp1.otp, otp_str1)
        self.assertEqual(otp2.otp, otp_str2)


class TestYubicoVerifySingle(unittest.TestCase):
    def setUp(self):
        yubico.API_URLS = ('127.0.0.1:8881/wsapi/2.0/verify',)
        yubico.DEFAULT_TIMEOUT = 2
        yubico.CA_CERTS_BUNDLE_PATH = None

        self.client_no_verify_sig = yubico.Yubico('1234', None,
                                                  use_https=False)
        self.client_verify_sig = yubico.Yubico('1234', 'secret123456',
                                               use_https=False)

    def test_invalid_custom_ca_certs_path(self):
        if hasattr(sys, 'pypy_version_info'):
            # TODO: Figure out why this breaks PyPy
            return

        yubico.CA_CERTS_BUNDLE_PATH = '/does/not/exist.1'
        client = yubico.Yubico('1234', 'secret123456')

        try:
            client.verify('bad')
        except requests.exceptions.SSLError:
            pass
        else:
            self.fail('SSL exception was not thrown')

    def test_replayed_otp(self):
        self._set_mock_action('REPLAYED_OTP')

        try:
            self.client_no_verify_sig.verify('bad')
        except StatusCodeError, e:
            self.assertEqual(e.status_code, 'REPLAYED_OTP')

    def test_verify_bad_status_codes(self):
        for status in (set(yubico.BAD_STATUS_CODES) - set(['REPLAYED_OTP'])):
            self._set_mock_action(status)

            try:
                self.client_no_verify_sig.verify('bad')
            except Exception, e:
                self.assertEqual(str(e), 'NO_VALID_ANSWERS')

    def test_verify_local_timeout(self):
        self._set_mock_action('timeout')

        try:
            self.client_no_verify_sig.verify('bad')
        except Exception, e:
            self.assertEqual(str(e), 'NO_VALID_ANSWERS')

    def test_verify_invalid_signature(self):
        self._set_mock_action('no_signature_ok')

        try:
            self.client_verify_sig.verify('test')
        except SignatureVerificationError:
            pass
        else:
            self.fail('Exception was not thrown')

    def test_verify_no_such_client(self):
        self._set_mock_action('no_such_client')

        try:
            self.client_no_verify_sig.verify('test')
        except InvalidClientIdError, e:
            self.assertEqual(e.client_id, '1234')
        else:
            self.fail('Exception was not thrown')

    def test_verify_ok_dont_check_signature(self):
        self._set_mock_action('no_signature_ok')

        status = self.client_no_verify_sig.verify('test')
        self.assertTrue(status)

    def test_verify_ok_check_signature(self):
        signature = \
            self.client_verify_sig.generate_message_signature('status=OK')
        self._set_mock_action('ok_signature', signature=signature)

        status = self.client_verify_sig.verify('test')
        self.assertTrue(status)

    def test_verify_invalid_otp_returned_in_the_response(self):
        self._set_mock_action('no_signature_ok_invalid_otp_in_response')

        try:
            self.client_no_verify_sig.verify('test')
        except InvalidValidationResponse, e:
            self.assertTrue('Unexpected OTP in response' in e.message)
        else:
            self.fail('Exception was not thrown')

    def test_verify_invalid_nonce_returned_in_the_response(self):
        self._set_mock_action('no_signature_ok_invalid_nonce_in_response')

        try:
            self.client_no_verify_sig.verify('test')
        except InvalidValidationResponse, e:
            self.assertTrue('Unexpected nonce in response' in e.message)
        else:
            self.fail('Exception was not thrown')

    def _set_mock_action(self, action, port=8881, signature=None):
        path = '/set_mock_action?action=%s' % (action)

        if signature:
            path += '&signature=%s' % (signature)

        requests.get(url='http://127.0.0.1:%s%s' % (port, path))


if __name__ == '__main__':
    sys.exit(unittest.main())



"""
Class which holds data about an OTP.
"""

import modhex


class OTP(object):
    def __init__(self, otp, translate_otp=True):
        self.otp = self.get_otp_modehex_interpretation(otp) \
            if translate_otp else otp

        self.device_id = self.otp[:12]
        self.session_counter = None
        self.timestamp = None
        self.session_user = None

    def get_otp_modehex_interpretation(self, otp):
        # We only use the first interpretation, because
        # if the OTP uses all 16 characters in its alphabet
        # there is only one possible interpretation of that otp
        try:
            interpretations = modhex.translate(unicode(otp))
        except Exception:
            return otp

        if len(interpretations) == 0:
            return otp
        elif len(interpretations) > 1:
            # If there are multiple interpretations first try to use the same
            # translation as the input OTP. If the one is not found, use the
            # random interpretation.
            if unicode(otp) in interpretations:
                return otp

        return interpretations.pop()

    def __repr__(self):
        return '%s, %s, %s' % (self.otp, self.device_id, self.timestamp)



# -*- coding: utf-8 -*-
#
# Name: Yubico Python Client
# Description: Python class for verifying Yubico One Time Passwords (OTPs).
#
# Author: Tomaž Muraus (http://www.tomaz.me)
# License: BSD
#
# Copyright (c) 2010, Tomaž Muraus
# Copyright (c) 2012, Yubico AB
# All rights reserved.
#
# Requirements:
# - Python >= 2.5

import re
import os
import time
import urllib
import hmac
import base64
import hashlib
import threading
import logging

import requests

from otp import OTP
from yubico_exceptions import (StatusCodeError, InvalidClientIdError,
                               InvalidValidationResponse,
                               SignatureVerificationError)

logger = logging.getLogger('yubico.client')

# Path to the custom CA certificates bundle. Only used if set.
CA_CERTS_BUNDLE_PATH = None

COMMON_CA_LOCATIONS = [
    '/usr/local/lib/ssl/certs/ca-certificates.crt',
    '/usr/local/ssl/certs/ca-certificates.crt',
    '/usr/local/share/curl/curl-ca-bundle.crt',
    '/usr/local/etc/openssl/cert.pem',
    '/opt/local/lib/ssl/certs/ca-certificates.crt',
    '/opt/local/ssl/certs/ca-certificates.crt',
    '/opt/local/share/curl/curl-ca-bundle.crt',
    '/opt/local/etc/openssl/cert.pem',
    '/usr/lib/ssl/certs/ca-certificates.crt',
    '/usr/ssl/certs/ca-certificates.crt',
    '/usr/share/curl/curl-ca-bundle.crt',
    '/etc/ssl/certs/ca-certificates.crt',
    '/etc/pki/tls/cert.pem',
    '/etc/pki/CA/cacert.pem',
    'C:\Windows\curl-ca-bundle.crt',
    'C:\Windows\ca-bundle.crt',
    'C:\Windows\cacert.pem'
]

API_URLS = ('api.yubico.com/wsapi/2.0/verify',
            'api2.yubico.com/wsapi/2.0/verify',
            'api3.yubico.com/wsapi/2.0/verify',
            'api4.yubico.com/wsapi/2.0/verify',
            'api5.yubico.com/wsapi/2.0/verify')

DEFAULT_TIMEOUT = 10            # How long to wait before the time out occurs
DEFAULT_MAX_TIME_WINDOW = 40    # How many seconds can pass between the first
                                # and last OTP generations so the OTP is
                                # still considered valid (only used in the
                                # multi mode) default is 5 seconds
                                # (40 / 0.125 = 5)

BAD_STATUS_CODES = ['BAD_OTP', 'REPLAYED_OTP', 'BAD_SIGNATURE',
                    'MISSING_PARAMETER', 'OPERATION_NOT_ALLOWED',
                    'BACKEND_ERROR', 'NOT_ENOUGH_ANSWERS',
                    'REPLAYED_REQUEST']


class Yubico(object):
    def __init__(self, client_id, key=None, use_https=True, verify_cert=True,
                 translate_otp=True):
        self.client_id = client_id
        self.key = base64.b64decode(key) if key is not None else None
        self.use_https = use_https
        self.verify_cert = verify_cert
        self.translate_otp = translate_otp

    def verify(self, otp, timestamp=False, sl=None, timeout=None,
               return_response=False):
        """
        Returns True is the provided OTP is valid,
        False if the REPLAYED_OTP status value is returned or the response
        message signature verification failed and None for the rest of the
        status values.
        """
        ca_bundle_path = self._get_ca_bundle_path()

        otp = OTP(otp, self.translate_otp)
        nonce = base64.b64encode(os.urandom(30), 'xz')[:25]
        query_string = self.generate_query_string(otp.otp, nonce, timestamp,
                                                  sl, timeout)
        request_urls = self.generate_request_urls()

        threads = []
        timeout = timeout or DEFAULT_TIMEOUT
        for url in request_urls:
            thread = URLThread('%s?%s' % (url, query_string), timeout,
                               self.verify_cert, ca_bundle_path)
            thread.start()
            threads.append(thread)

        # Wait for a first positive or negative response
        start_time = time.time()
        while threads and (start_time + timeout) > time.time():
            for thread in threads:
                if not thread.is_alive():
                    if thread.exception:
                        raise thread.exception
                    elif thread.response:
                        status = self.verify_response(thread.response,
                                                      otp.otp, nonce,
                                                      return_response)

                        if status:
                            if return_response:
                                return status
                            else:
                                return True
                    threads.remove(thread)
            time.sleep(0.1)

        # Timeout or no valid response received
        raise Exception('NO_VALID_ANSWERS')

    def verify_multi(self, otp_list=None, max_time_window=None, sl=None,
                     timeout=None):
        # Create the OTP objects
        otps = []
        for otp in otp_list:
            otps.append(OTP(otp, self.translate_otp))

        device_ids = set()
        for otp in otps:
            device_ids.add(otp.device_id)

        # Check that all the OTPs contain same device id
        if len(device_ids) != 1:
            raise Exception('OTPs contain different device ids')

        # Now we verify the OTPs and save the server response for each OTP.
        # We need the server response, to retrieve the timestamp.
        # It's possible to retrieve this value locally, without querying the
        # server but in this case, user would need to provide his AES key.
        for otp in otps:
            response = self.verify(otp.otp, True, sl, timeout,
                                   return_response=True)

            if not response:
                return False

            otp.timestamp = int(response['timestamp'])

        count = len(otps)
        delta = otps[count - 1].timestamp - otps[0].timestamp

        if max_time_window:
            max_time_window = (max_time_window / 0.125)
        else:
            max_time_window = DEFAULT_MAX_TIME_WINDOW

        if delta > max_time_window:
            raise Exception('More then %s seconds has passed between ' +
                            'generating the first and the last OTP.' %
                            (max_time_window * 0.125))

        return True

    def verify_response(self, response, otp, nonce, return_response=False):
        """
        Returns True if the OTP is valid (status=OK) and return_response=False,
        otherwise (return_response = True) it returns the server response as a
        dictionary.

        Throws an exception if the OTP is replayed, the server response message
        verification failed or the client id is invalid, returns False
        otherwise.
        """
        try:
            status = re.search(r'status=([A-Z0-9_]+)', response) \
                       .groups()

            if len(status) > 1:
                message = 'More than one status= returned. Possible attack!'
                raise InvalidValidationResponse(message, response)

            status = status[0]
        except (AttributeError, IndexError):
            return False

        signature, parameters = \
            self.parse_parameters_from_response(response)

        # Secret key is specified, so we verify the response message
        # signature
        if self.key:
            generated_signature = \
                self.generate_message_signature(parameters)

            # Signature located in the response does not match the one we
            # have generated
            if signature != generated_signature:
                raise SignatureVerificationError(generated_signature,
                                                 signature)
        param_dict = self.get_parameters_as_dictionary(parameters)

        if 'otp' in param_dict and param_dict['otp'] != otp:
            message = 'Unexpected OTP in response. Possible attack!'
            raise InvalidValidationResponse(message, response, param_dict)

        if 'nonce' in param_dict and param_dict['nonce'] != nonce:
            message = 'Unexpected nonce in response. Possible attack!'
            raise InvalidValidationResponse(message, response, param_dict)

        if status == 'OK':
            if return_response:
                return param_dict
            else:
                return True
        elif status == 'NO_SUCH_CLIENT':
            raise InvalidClientIdError(self.client_id)
        elif status == 'REPLAYED_OTP':
            raise StatusCodeError(status)

        return False

    def generate_query_string(self, otp, nonce, timestamp=False, sl=None,
                              timeout=None):
        """
        Returns a query string which is sent to the validation servers.
        """
        data = [('id', self.client_id),
                ('otp', otp),
                ('nonce', nonce)]

        if timestamp:
            data.append(('timestamp', '1'))

        if sl is not None:
            if sl not in range(0, 101) and sl not in ['fast', 'secure']:
                raise Exception('sl parameter value must be between 0 and '
                                '100 or string "fast" or "secure"')

            data.append(('sl', sl))

        if timeout:
            data.append(('timeout', timeout))

        query_string = urllib.urlencode(data)

        if self.key:
            hmac_signature = self.generate_message_signature(query_string)
            query_string += '&h=%s' % (hmac_signature.replace('+', '%2B'))

        return query_string

    def generate_message_signature(self, query_string):
        """
        Returns a HMAC-SHA-1 signature for the given query string.
        http://goo.gl/R4O0E
        """
        pairs = query_string.split('&')
        pairs = [pair.split('=') for pair in pairs]
        pairs_sorted = sorted(pairs)
        pairs_string = '&' . join(['=' . join(pair) for pair in pairs_sorted])

        digest = hmac.new(self.key, pairs_string, hashlib.sha1).digest()
        signature = base64.b64encode(digest)

        return signature

    def parse_parameters_from_response(self, response):
        """
        Returns a response signature and query string generated from the
        server response. 'h' aka signature argument is stripped from the
        returned query string.
        """
        split = [pair.strip() for pair in response.split('\n')
                 if pair.strip() != '']
        query_string = '&' . join(split)
        split_dict = self.get_parameters_as_dictionary(query_string)

        if 'h' in split_dict:
            signature = split_dict['h']
            del split_dict['h']
        else:
            signature = None

        query_string = ''
        for index, (key, value) in enumerate(split_dict.iteritems()):
            query_string += '%s=%s' % (key, value)

            if index != len(split_dict) - 1:
                query_string += '&'

        return (signature, query_string)

    def get_parameters_as_dictionary(self, query_string):
        """ Returns query string parameters as a dictionary. """
        dictionary = dict([parameter.split('=', 1) for parameter
                           in query_string.split('&')])

        return dictionary

    def generate_request_urls(self):
        """
        Returns a list of the API URLs.
        """
        urls = []
        for url in API_URLS:
            if self.use_https:
                url = 'https://%s' % (url)
            else:
                url = 'http://%s' % (url)
            urls.append(url)

        return urls

    def _get_ca_bundle_path(self):
        """
        Return a path to the CA bundle which is used for verifying the hosts
        SSL certificate.
        """
        if CA_CERTS_BUNDLE_PATH:
            # User provided a custom path
            return CA_CERTS_BUNDLE_PATH

        for file_path in COMMON_CA_LOCATIONS:
            if os.path.exists(file_path) and os.path.isfile(file_path):
                return file_path

        return None


class URLThread(threading.Thread):
    def __init__(self, url, timeout, verify_cert, ca_bundle_path=None):
        super(URLThread, self).__init__()
        self.url = url
        self.timeout = timeout
        self.verify_cert = verify_cert
        self.ca_bundle_path = ca_bundle_path
        self.exception = None
        self.request = None
        self.response = None

    def run(self):
        logger.debug('Sending HTTP request to %s (thread=%s)' % (self.url,
                                                                 self.name))
        verify = self.verify_cert

        if self.ca_bundle_path is not None:
            verify = self.ca_bundle_path
            logger.debug('Using custom CA bunde: %s' % (self.ca_bundle_path))

        try:
            self.request = requests.get(url=self.url, timeout=self.timeout,
                                        verify=verify)
            self.response = self.request.content
        except requests.exceptions.SSLError, e:
            self.exception = e
            self.response = None
        except Exception, e:
            logger.error('Failed to retrieve response: ' + str(e))
            self.response = None

        args = (self.url, self.name, self.response)
        logger.debug('Received response from %s (thread=%s): %s' % args)
